# -*- coding: utf-8 -*-
'''
SmileDecoder: decode smile dataformat input to python object
(dictionary, array or value)
Typical usage:
    from decoder import SmileDecoder
    with open(smile_path, 'rb') as smile_file:
            smile_contents = smile_file.read()
    decoder = SmileDecoder()
    decoded_data = decoder.decode(smile_contents)
'''
import struct

# too few public methods: ACQ
# pylint: disable=R0903
class SmileDecoder:
    '''
    Smile decoding methods
    '''
    def __init__(self, encoding='utf-8'):
        '''
        Class initialization
        Args:
            encoding (string): encoding used for decoding unicode strings
        '''
        self._properties = {
            'raw-binary': False,
            'shared-keys-names': False,
            'shared-string-values': True
        }
        self._encoding = encoding
        self._buffer = None
        self._current_obj = None
        self._index = 0
        self._shared_keys = []
        self._shared_values = []

    def _set_shared_key(self, key):
        '''
        Save shared key in dedicated buffer
        '''
        if len(key) <= 64:
            if len(self._shared_keys) >= 1024:
                self._shared_keys = []
            self._shared_keys.append(key)

    def _set_shared_value(self, value):
        '''
        Save shared value in dedicated buffer
        '''
        if len(value) <= 64:
            if len(self._shared_values) >= 1024:
                self._shared_values = []
            self._shared_values.append(value)

    @classmethod
    def _unzigzag(cls, value, mask = None):
        '''
        Unzigzag zigzag encoded int value
        '''
        if mask:
            value = value & mask
        return (value >> 1) ^ -(value & 1)

    def _float(self, bytes_count):
        '''
        Single or Double precision float decoding
        '''
        value = 0
        fmt = {4: '>f', 8: '>d'}[bytes_count]
        tokens = self._get_bytes(bytes_count + bytes_count // 4)
        for token in tokens if bytes_count == 8 else reversed(tokens):
            value <<= 7
            value |= token
        return struct.unpack(fmt, value.to_bytes(bytes_count, byteorder='big'))[0]

    def _int(self, unzigzag=True):
        '''
        Variable length int decoding
        '''
        value = 0
        token = self._get_bytes(1)[0]
        while token & 2**7 == 0:
            value <<= 7
            value |= token & 2**7-1
            token = self._get_bytes(1)[0]
        value <<= 6
        value |= token & 2**6-1
        return self._unzigzag(value, None) if unzigzag else value

    def _bigint(self):
        '''
        Big integer decoding
        '''
        length = self._int(unzigzag=False)
        remain = 8 * length % 7
        bytes_count = int((8 * length - remain)/7)
        value = 0
        tokens = self._get_bytes(bytes_count)
        for token in tokens:
            value <<= 7
            value |= token
        if remain > 0:
            value <<= remain
            value |= self._get_bytes(1)[0] & 0x7f
        return value

    def _string(self, token_type, length, encoding):
        '''
        String decoding
        '''
        if length:
            # fixed length string
            buffer = self._get_bytes(length).decode(encoding)
            if token_type == 'value':
                self._set_shared_value(buffer)
            elif token_type == 'key':
                self._set_shared_key(buffer)
        else:
            # Variable length string
            buffer = bytearray()
            token = self._get_bytes(1)[0]
            while token != 0xfc:
                buffer.append(token)
                token = self._get_bytes(1)[0]
            buffer = buffer.decode(encoding)
        return buffer

    def _get_bytes(self, count):
        '''
        Get bytes from buffer and increase index
        '''
        buffer = self._buffer[self._index: self._index + count]
        self._index += count
        return buffer

    def _decode_object(self):
        '''
        Entry point for decoding dictionary
        '''
        obj = {}
        while self._buffer[self._index] != 0xfb:
            key = self._decode_key()
            value = self._decode_value()
            obj[key] = value
        self._index += 1
        return dict(sorted(obj.items()))

    def _decode_array(self):
        '''
        Entry point for decoding array
        '''
        array = []
        while self._buffer[self._index] != 0xf9:
            value = self._decode_value()
            array.append(value)
        self._index += 1
        return array

    def _get_referenced(self, token_type, reference):
        '''
        Retrieve referenced key or value in buffer
        '''
        if (token_type == 'value' and not self._properties['shared-string-values']) or \
            (token_type == 'key' and not self._properties['shared-keys-names']):
            raise ValueError(f'Shared References not setted for {token_type}s')
        if token_type == 'value':
            referenced =  self._shared_values[reference]
        else:
            referenced =  self._shared_keys[reference]
        return referenced

    def _decode_value(self):
        '''
        Decoding a value (dispatch for int, float, string, array or dict)
        '''
        value = None
        token = self._get_bytes(1)[0]
        if 0x00 <= token <= 0x1f:
            # Short value String Reference
            value = self._get_referenced('value', token - 1)
        elif 0x20 <= token <= 0x23:
            # Special values: empty string, None, False, True
            value = {0x20: "", 0x21: None, 0x22: False, 0x23: True}[token]
        elif 0x24 <= token <= 0x25:
            # Integer
            value = self._int()
        elif token == 0x26:
            # Big Integer
            value = self._bigint()
        elif token == 0x27:
            # Unused -> Exception
            raise ValueError(f'Value : {token:02x}:  reserved for future use')
        elif token == 0x28:
            # Float (32 bits)
            value = self._float(4)
        elif token == 0x29:
            # Double (64 bits)
            value = self._float(8)
        elif 0x40 <= token <= 0x5f:
            # Tiny ASCII
            value = self._string('value', (token & 2**5-1) + 1, 'ascii')
        elif 0x60 <= token <= 0x7f:
            # Short ASCII
            value = self._string('value', (token & 2**5-1) + 33, 'ascii')
        elif 0x80 <= token <= 0x9f:
            # Tiny UNICODE
            value = self._string('value', (token & 2**5-1) + 2, self._encoding)
        elif 0xa0 <= token <= 0xbf:
            # Short UNICODE
            value = self._string('value', (token & 2**5-1) + 34, self._encoding)
        elif 0xc0 <= token <= 0xdf:
            # Small Integer
            value = SmileDecoder._unzigzag(token, 2**5-1)
        elif token == 0xe0:
            # Non referencable ASCII text
            value = self._string('value', None, 'ascii')
        elif token == 0xe4:
            # Non referencable UNICODE text
            value = self._string('value', None, self._encoding)
        elif 0xec <= token < 0xf0:
            # Long value String Reference
            value = self._get_referenced('value', (token & 0x3) << 8 | self._get_bytes(1)[0])
        elif token == 0xfa:
            # Sub Object
            value = self._decode_object()
        elif token == 0xf8:
            # Sub Array
            value = self._decode_array()
        elif token == 0xfd:
            # binary (raw)
            length = self._int(False)
            value = self._get_bytes(length)
        else:
            raise ValueError(f'value: {token:02x}: unknown token')
        return value

    def _decode_key(self):
        '''
        Decoding dict key
        '''
        key = None
        token = self._get_bytes(1)[0]
        # reserved
        if (0 <= token <= 0x1f) or \
            (0x21 <= token <= 0x2f):
            raise ValueError(f'key: {self._buffer[self._index]:02x}: reserved for future use')
        # empty string
        elif token == 0x20:
            key = ""
        # Long Shared Key Name Reference
        elif 0x30 <= token <= 0x33:
            key = self._get_referenced('key', (token & 0b11) << 8 | self._get_bytes(1)[0])
        elif token == 0x34:
        # Long NOT Shared Unicode Key Name
            self._string('key', None, self._encoding)
        # Short Shared Key Name Reference
        elif 0x40 <= token <= 0x7F:
            key = self._get_referenced('key', token - 0x40)
        # Short ASCII Name
        elif 0x80 <= token <= 0xBF:
            key = self._string('key', token - 0x80 + 1, 'ascii')
        # Short unicode name
        elif 0xC0 <= token <= 0xf7:
            key = self._string('key', token - 0xc0 + 2, self._encoding)
        else:
            raise ValueError(f'key error: {self._buffer[self._index]:02x}')
        return key

    def _decode_header(self):
        '''
        Decoding header
        '''
        buffer = self._get_bytes(4)
        if buffer[:3].decode(self._encoding) != ':)\n':
            raise ValueError('invalid header')
        self._properties['raw-binary'] = buffer[3] & 4 > 0
        self._properties['shared-keys-names'] = buffer[3] & 2 > 0
        self._properties['shared-string-values'] = buffer[3] & 1 > 0

    def decode(self, buffer):
        '''
        Class main entry point
        Args:
            buffer (bytes | bytearray | list of bytes): smile data to decode
        Typical usage:
            with open(smile_path, 'rb') as smile_file:
                smile_contents = smile_file.read()
            decoder = SmileDecoder()
            decoded_data = decoder.decode(smile_contents)
        '''
        self._index = 0
        self._shared_keys = []
        self._shared_values = []
        if not isinstance(buffer, (bytes, bytearray)):
            self._buffer = bytes([ord(byte) for byte in buffer])
        else:
            self._buffer = buffer
        self._decode_header()
        return self._decode_value()



# -*- coding: utf-8 -*-
'''
Automatic tests for SmileDecoder and SmileEncoder
'''
import json
from pathlib import Path
from newsmile import SmileDecoder
from newsmile import SmileEncoder

# disable 'too few public methods' warning
# pylint: disable=R0903
class ObjectComparator:
    '''
    Utility class for compare two objects
    checks:
        if dictionaries have same keys and same values (recursive)
        if arrays contains same values
        if float values are the same (+/- eps)
    '''
    def __init__(self):
        self._current = None

    def compare(self, objecta, objectb):
        '''
        Main method for ObjectComparator class
        Args:
            objecta (any): first object to compare
            objectb (any): second object to compare
        '''
        flag = True
        if isinstance(objecta, dict) and isinstance(objectb, dict):
            self._current = objecta, objectb
            for key in objecta.keys():
                if key in objectb:
                    flag = self.compare(objecta[key], objectb[key])
                    if not flag:
                        break
                else:
                    break
        elif isinstance(objecta, list) and isinstance(objectb, list):
            for idx, item in enumerate(objecta):
                if item in objectb:
                    flag = self.compare(objecta[idx], objectb[objectb.index(item)])
                    if not flag:
                        break
        else:
            if isinstance(objecta, float) and isinstance(objectb, float):
                flag = abs(objecta - objectb) < 2.220446049250313e-16
            else:
                flag = objecta == objectb
        return flag

if __name__ == '__main__':
    print('### Decoding Test ###')
    decoder = SmileDecoder()
    for smile_path in Path('data-test/smile').iterdir():
        json_path = f'data-test/json/{smile_path.stem}.jsn'
        with open(smile_path, 'rb') as smile_file:
            from_smile = decoder.decode(smile_file.read())
        with open(json_path, 'r', encoding='utf-8') as json_file:
            from_json = json.loads(json_file.read())
        RESULT = 'passed' if ObjectComparator().compare(from_smile, from_json) else 'failed'
        print(f"{smile_path.stem:20s}: {RESULT}")

    print('### Encoding Test ###')
    encoder = SmileEncoder(shared_values=True, single_or_double=4)
    for json_path in Path('data-test/json').iterdir():
        smile_path = f'data-test/smile/{json_path.stem}.smile'
        with open(json_path, 'r', encoding='utf-8') as json_file:
            encoded = encoder.encode(json_file.read())
        with open(smile_path, 'rb') as smile_file:
            from_smile = decoder.decode(smile_file.read())
        from_encoded = decoder.decode(encoded)
        RESULT = 'passed' if ObjectComparator().compare(from_smile, from_encoded) else 'failed'
        print(f"{json_path.stem:20s}: {RESULT}")



# -*- coding: utf-8 -*-
'''
SmileEncoder: encode an object to smile format
'''
import math
import json
import struct

# too few public methods: ACQ
# pylint: disable=R0903
# too many attributes: ACQ
# pylint: disable=R0902
class SmileEncoder:
    '''
    Smile encoding methods encapsulation
    '''
    # too many arguments: ACQ
    # pylint: disable=R0913
    def __init__(self,
        raw_binary=False,
        shared_values=False,
        shared_keys=True,
        single_or_double=0,
        encoding='utf-8'):
        '''
        Args:
            raw_binary (bool): contains binary data
            shared_values (bool): maintains shared values buffer and references
            shared_keys (bool): maintains shared keys buffer and references
            single_or_double (0|4|8): float values precisions (0=auto)
            encoding (string): encoding used for unicode values and keys
        '''
        self._encoding = encoding
        # prepare header's bytes from parameters
        self._buffer = bytearray()
        self._raw_binary = raw_binary
        self._shared_values = shared_values
        self._shared_keys = shared_keys
        # 0: auto, 4: single, 8: double
        self._single_or_double = single_or_double
        self._shared_value_buffer = []
        self._shared_keys_buffer = []

    def encode(self, data):
        '''
        Start encoding: top-level function
        Args:
            data json data string to encode
        '''
        object_to_encode = json.loads(data)
        # reinit buffers
        self._buffer = bytearray()
        self._shared_value_buffer = []
        self._shared_keys_buffer = []
        # write header
        self._set_header()
        if isinstance(object_to_encode, dict):
            self._encode_object(object_to_encode)
        elif isinstance(object_to_encode, list):
            self._encode_array(object_to_encode)
        else:
            self._encode_value(object_to_encode)
        return bytes(self._buffer)

    def _is_ascii(self, value):
        '''
        Check if string is ascii or unicode
        '''
        is_ascii, buffer = True, None
        # in_ascii = not (True in [ord(character) > 127 for character in value])
        for character in value:
            if ord(character) > 127:
                is_ascii = False
                break
        if is_ascii:
            buffer = value.encode('ascii')
        else:
            buffer = value.encode(self._encoding)
        return is_ascii, buffer

    @classmethod
    def _zigzag(cls, value, nbits=5):
        '''
        ZigZag encode value on nbits
        '''
        return (value >> (nbits - 1)) ^ (value << 1)

    def _set_header(self):
        '''
        Write header to buffer
        '''
        self._buffer.extend(':)\n\x00'.encode(self._encoding))
        if self._raw_binary:
            self._buffer[-1] |= 0x4
        if self._shared_values:
            self._buffer[-1] |= 0x2
        if self._shared_keys:
            self._buffer[-1] |= 0x1

    def _float_or_double(self, value):
        '''
        Check if value is single or double precision
        (dummy check: simply check if string representation
         ends with 000)
        '''
        if self._single_or_double == 0:
            buf = f'{value:.15f}'[-3:] == '000'
            return 4 if buf else 8
        return self._single_or_double

    def _encode_object(self, dico):
        '''
        Encode a dict
        '''
        if not isinstance(dico, dict):
            raise ValueError('Value should be a dictionnary')
        self._buffer.append(0xfa)
        for key in sorted(dico.keys()):
            self._encode_key(key)
            self._encode_value(dico[key])
        self._buffer.append(0xfb)

    def _encode_array(self, iterable):
        '''
        Encode an array (list or tuple)
        '''
        if not isinstance(iterable, (list, tuple)):
            raise ValueError('Value should be an iterable')
        self._buffer.append(0xf8)
        for value in iterable:
            self._encode_value(value)
        self._buffer.append(0xf9)

    def _encode_vint(self, value, nbits=None, zigzag=True):
        '''
        Variable length int encoding
        '''
        if zigzag and nbits:
            value = self._zigzag(value, nbits=nbits)
        buffer = [(value & 2**6-1) | 2**7]
        processed_bits = 6
        value >>= 6
        while value != 0 and (nbits is None or processed_bits < nbits):
            buffer.append(value & 2**7-1)
            value >>= 7
            processed_bits += 7
        self._buffer.extend(reversed(buffer))

    def _encode_int(self, value):
        '''
        Encoding integers
        '''
        if -16 <= value <= 15:
            self._buffer.append(0xc0 | self._zigzag(value, 5))
        elif -2 ** 31 <= value <= 2 ** 31 - 1:
            self._buffer.append(0x24)
            self._encode_vint(value, 32)
        elif -2 ** 63 <= value <= 2 ** 63 - 1:
            self._buffer.append(0x25)
            self._encode_vint(value, 64)
        else:
            # Big Integer
            bytes_count = math.ceil(value.bit_length() / 8)
            buffer = []
            shift = value.bit_length() - (value.bit_length() % 8 - 1)
            while shift > 0:
                buffer.append((value & (0x7f << shift)) >> shift)
                shift -= 7
            if shift != 0:
                buffer.append((value & 2 ** (7 + shift) - 1))
            self._buffer.append(0x26)
            self._encode_vint(bytes_count)
            self._buffer.extend(buffer)

    def _encode_float(self, value):
        '''
        Single or Double precision encoding
        '''
        nbytes = self._float_or_double(value)
        self._buffer.append({4: 0x28, 8: 0x29}[nbytes])
        buffer = struct.pack({4: '>f', 8: '>d'}[nbytes], value)
        buffer = struct.unpack({4: '>I', 8: '>Q'}[nbytes], buffer)[0]
        tmp_buf = []
        for _ in range({4: 5, 8: 10}[nbytes]):
            tmp_buf.append(buffer & 0x7f)
            buffer >>= 7
        self._buffer.extend(tmp_buf if nbytes == 4 else reversed(tmp_buf))

    def _encode_string_value(self, value):
        '''
        Encodage d'une valeur chaine
        '''
        if value == '':
            self._buffer.append(0x20)
            return
        is_ascii, buffer = self._is_ascii(value)
        if value in self._shared_value_buffer:
            # referenced value
            idx = self._shared_value_buffer.index(value)
            if idx < 31:
                # short index
                self._buffer.append(0x00 | (idx + 1 & 0x1f))
            else:
                # long index
                self._buffer.append(0xec | (idx & 0x300) >> 8)
                self._buffer.append(idx & 0xff)
        else:
            if len(buffer) <= (32 if is_ascii else 33):
                # tiny ascii or unicode
                self._buffer.append((0x40 if is_ascii else 0x80) |
                    (len(buffer) - (1 if is_ascii else 2)) & 0x1f)
                self._buffer.extend(buffer)
            elif len(buffer) <= (64 if is_ascii else 65):
                # short ascii or unicode
                self._buffer.append((0x60 if is_ascii else 0xa0) |
                    ((len(buffer) - (33 if is_ascii else 34)) & 0x1f))
                self._buffer.extend(buffer)
            elif len(buffer) > (64 if is_ascii else 65):
                # non referencable value
                self._buffer.append(0xe0 if is_ascii else 0xe4)
                self._buffer.extend(buffer)
                self._buffer.append(0xfc)
            if len(value) < 64 and self._shared_values:
                # cas d'une chaine elligible au buffer
                if len(self._shared_value_buffer) >= 1024:
                    self._shared_value_buffer = []
                self._shared_value_buffer.append(value)

    def _encode_string_key(self, key):
        '''
        Encoding a string key
        '''
        is_ascii, buffer = self._is_ascii(key)
        if key in self._shared_keys_buffer:
            # referenced key
            idx = self._shared_keys_buffer.index(key)
            if idx < 64:
                # short index
                self._buffer.append(0x40 | idx)
            else:
                # long index
                self._buffer.append(0x30 | idx >> 8)
                self._buffer.append(idx & 0xff)
        else:
            if len(buffer) <= 64:
                # fixed length string
                self._buffer.append((0x80 if is_ascii else 0xc0) | (len(key) - 1))
                self._buffer.extend(buffer)
            else:
                # variable length string
                self._buffer.append(0x34)
                self._buffer.extend(buffer)
                self._buffer.append(0xfc)
            if len(key) < 64 and self._shared_keys:
                # elligible string
                if len(self._shared_keys_buffer) >= 1024:
                    self._shared_keys_buffer = []
                self._shared_keys_buffer.append(key)

    def _encode_key(self, key):
        '''
        Top-level key encoding
        '''
        if isinstance(key, str):
            if key == "":
                # empty string
                self._buffer.append(0x20)
            else:
                self._encode_string_key(key)
        else:
            raise ValueError('Key case!')

    def _encode_value(self, value):
        '''
        Top-level value encoding
        '''
        if value == "":
            self._buffer.append(0x20)
        elif value is None:
            self._buffer.append(0x21)
        elif value is False:
            self._buffer.append(0x22)
        elif value is True:
            self._buffer.append(0x23)
        elif isinstance(value, int): # 0x20 -> 0x3f | 0xc0 -> 0xdf
            self._encode_int(value)
        elif isinstance(value, float):
            self._encode_float(value)
        elif isinstance(value, str): # 0x00 -> 0x1f / 0x40 -> 0xbf / 0xe0 -> 0xff
            self._encode_string_value(value)
        elif isinstance(value, dict):
            self._encode_object(value)
        elif isinstance(value, (list, tuple)):
            self._encode_array(value)
        else:
            raise ValueError('Value case!')
